#include  <chip.h>
#include  <us.h>
#include  <dxclient.h>
#include  <piocfg.h>
#include  <wl_delay.h>

#include  <FreeRTOS.h>
#include  <task.h>

//#define _USE_MYTASK

#define _TXD  16
#define _RXD  26
#define _RTS  27
#define _LED  24

//-----------------------------
// コントロールテーブル
//-----------------------------
typedef struct {
  uint16_t    ModelNumber;
  uint8_t     FirmwareVersion;
  uint8_t     ID;
  uint8_t     Baudrate;
  uint8_t     WriteNVM;
  uint8_t     LED;
  uint8_t     PinCfg[12];
  uint8_t     BaseClock;
  struct {
    uint16_t  cycle;
    uint16_t  duty;
  } __attribute__((packed)) PWMValue[6];
  uint16_t    capture[4];
  uint16_t    OUT;
  uint16_t    IN;
  uint16_t    ADV[12];
  uint8_t     userarea[16];
} __attribute__((packed)) TE097CtrlTable;

#define _ADDR_PinCfg      ((int)&((TE097CtrlTable*)0)->PinCfg)
#define _ADDR_BaseClock   ((int)&((TE097CtrlTable*)0)->BaseClock)
#define _ADDR_PWMValue    ((int)&((TE097CtrlTable*)0)->PWMValue)
#define _ADDR_PWMcapture  ((int)&((TE097CtrlTable*)0)->capture)
#define _ADDR_OUT         ((int)&((TE097CtrlTable*)0)->OUT)
#define _ADDR_IN          ((int)&((TE097CtrlTable*)0)->IN)
#define _ADDR_ADV         ((int)&((TE097CtrlTable*)0)->ADV)
#define _ADDR_userarea    ((int)&((TE097CtrlTable*)0)->userarea)

//-----------------------------
// コントロールテーブルデフォルト値
//-----------------------------
static const TE097CtrlTable DEFAULT_DATA = {
  ModelNumber     :0x4020,
  FirmwareVersion :'A',
  ID              :200,
  Baudrate        :1,
  WriteNVM        :0,
};

// Pin Mode
typedef enum {tPinDIN = 0, tPinDIN_PU, tPinDIN_PD, tPinDOUT, tPinDOUT_OC, tPinAIN, tPinPWM0, tPinPWM1, tPinPWM2, tPinPWM3, tPinPWM4, tPinPWM5, tPinMPW0, tPinMPW1, tPinMPW2, tPinMPW3, tPinMPC0, tPinMPC1, tPinMPC2, tPinMPC3, tPinSPFUNC} TMIOPinMode;

// Pin information
typedef struct {
  uint8_t               No;
  CHIP_PINx_T           CPin;
  CHIP_SWM_PIN_FIXED_T  AD;
} TPinInfo;

// 端子機能
static const TPinInfo PinInfo[12] = {
  {7,  IOCON_PIO7,  SWM_FIXED_ADC0},
  {6,  IOCON_PIO6,  SWM_FIXED_ADC1},
  {14, IOCON_PIO14, SWM_FIXED_ADC2},
  {23, IOCON_PIO23, SWM_FIXED_ADC3},
  {22, IOCON_PIO22, SWM_FIXED_ADC4},
  {21, IOCON_PIO21, SWM_FIXED_ADC5},
  {20, IOCON_PIO20, SWM_FIXED_ADC6},
  {19, IOCON_PIO19, SWM_FIXED_ADC7},
  {18, IOCON_PIO18, SWM_FIXED_ADC8},
  {17, IOCON_PIO17, SWM_FIXED_ADC9},
  {13, IOCON_PIO13, SWM_FIXED_ADC10},
  {4,  IOCON_PIO4,  SWM_FIXED_ADC11}
};

static const TPin pins[] = {
  {_TXD,  PIO_TYPE_MOVABLE,  SWM_U0_TXD_O, PIO_MODE_DEFAULT },
  {_RXD,  PIO_TYPE_MOVABLE,  SWM_U0_RXD_I, PIN_MODE_PULLUP  },
  {_RTS,  PIO_TYPE_MOVABLE,  SWM_U0_RTS_O, PIO_MODE_DEFAULT },
  {_LED,  PIO_TYPE_OUTPUT_1, 0,            PIO_MODE_DEFAULT },
};

// シリアル通信バッファ
uint8_t txd[50], rxd[150];

/*======================================================================*/
/* ETC                                                                  */
/*======================================================================*/
//-----------------------------
// LED
//-----------------------------
static void ToggleLED(void) {
  Chip_GPIO_SetPinToggle(LPC_GPIO_PORT, 0, _LED);
}

static void SetLED(bool On) {
  Chip_GPIO_SetPinState(LPC_GPIO_PORT, 0, _LED, (bool)!On);
}

//-----------------------------
// pin configuration
//-----------------------------
static void pin_config(uint8_t adch, TMIOPinMode pm) {
  const uint8_t pinno[] = { 7, 6, 14, 23, 22, 21, 20, 19, 18, 17, 13, 4 };
  const uint8_t fixedad[] = { SWM_FIXED_ADC0, SWM_FIXED_ADC1, SWM_FIXED_ADC2, SWM_FIXED_ADC3, SWM_FIXED_ADC4, SWM_FIXED_ADC5, SWM_FIXED_ADC6, SWM_FIXED_ADC7, SWM_FIXED_ADC8, SWM_FIXED_ADC9, SWM_FIXED_ADC10, SWM_FIXED_ADC11 };
  TPin pin;

//  Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_SWM);

//  Chip_SWM_DisableFixedPin(PinInfo[adch].AD);
//  Chip_IOCON_PinSetMode(LPC_IOCON, PinInfo[adch].CPin, PIN_MODE_INACTIVE);

  pin.gpio_no = pinno[adch];
  switch(pm) {
    case tPinDIN:
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDIN_PU:
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLUP;
      break;
    case tPinDIN_PD:
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLDOWN;
      break;
    case tPinDOUT:
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDOUT_OC:
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAIN:
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = fixedad[adch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinPWM0:
      break;
    case tPinPWM1:
      break;
    case tPinPWM2:
      break;
    case tPinPWM3:
      break;
    case tPinPWM4:
      break;
    case tPinPWM5:
      break;
    case tPinMPW0:
      break;
    case tPinMPW1:
      break;
    case tPinMPW2:
      break;
    case tPinMPW3:
      break;
    case tPinMPC0:
      break;
    case tPinMPC1:
      break;
    case tPinMPC2:
      break;
    case tPinMPC3:
      break;
    case tPinSPFUNC:
      break;
  }
  PIO_Configure (&pin, 1);
}

//-----------------------------
// Pin config
//-----------------------------
static bool SetPinAssign(CHIP_SWM_PIN_MOVABLE_T movable, int pin) {
  if((pin < 12) && (movable != SWM_U0_TXD_O) && (movable != SWM_U0_RXD_I) && (movable != SWM_U0_RTS_O)) {
    Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_SWM);
    Chip_SWM_DisableFixedPin(PinInfo[pin].AD);
    Chip_SWM_MovablePinAssign(movable, PinInfo[pin].No);
    Chip_Clock_DisablePeriphClock(SYSCTL_CLOCK_SWM);
    return true;
  }
  return false;
}

//-----------------------------
// WDT
//-----------------------------
static void init_wdt(void) {
	uint32_t wdtFreq;
  Chip_Clock_SetWDTOSC(WDTLFO_OSC_0_60, 64);      // Freq = 0.6Mhz, divided by 64. WDT_OSC should be 9.375khz
  Chip_SYSCTL_PowerUp(SYSCTL_SLPWAKE_WDTOSC_PD);  // Enable the power to the WDT
  wdtFreq = Chip_Clock_GetWDTOSCRate() / 4;       // The WDT divides the input frequency into it by 4
  Chip_WWDT_Init(LPC_WWDT);                       // Initialize WWDT (also enables WWDT clock)
  /* Set watchdog feed time constant to approximately 2s
     Set watchdog warning time to 512 ticks after feed time constant
     Set watchdog window time to 3s */
  Chip_WWDT_SetTimeOut(LPC_WWDT, wdtFreq * 2);
  Chip_WWDT_SetWarning(LPC_WWDT, 512);
  Chip_WWDT_SetWindow(LPC_WWDT, wdtFreq * 3);
  Chip_WWDT_SetOption(LPC_WWDT, WWDT_WDMOD_WDRESET);  // Configure WWDT to reset on timeout
  Chip_WWDT_ClearStatusFlag(LPC_WWDT, WWDT_WDMOD_WDTOF | WWDT_WDMOD_WDINT); // Clear watchdog warning and timeout interrupts
  Chip_WWDT_Start(LPC_WWDT);  // Start watchdog
}

//-----------------------------
// ADC
//-----------------------------
static void init_adc(void) {
  // Setup ADC for 12-bit mode and normal power
  Chip_ADC_Init(LPC_ADC, 0);
  // Need to do a calibration after initialization and trim
  Chip_ADC_StartCalibration(LPC_ADC);
  while(!(Chip_ADC_IsCalibrationDone(LPC_ADC))) ;
  // Setup for maximum ADC clock rate using sycnchronous clocking
  Chip_ADC_SetClockRate(LPC_ADC, 10000);  // 10kHz
  // Setup a sequencer to do the following
  Chip_ADC_SetupSequencer(LPC_ADC, ADC_SEQA_IDX, (ADC_SEQ_CTRL_CHANSEL(0) | ADC_SEQ_CTRL_CHANSEL(1) | ADC_SEQ_CTRL_CHANSEL(2) | ADC_SEQ_CTRL_CHANSEL(3) | ADC_SEQ_CTRL_CHANSEL(4) | ADC_SEQ_CTRL_CHANSEL(5) | ADC_SEQ_CTRL_CHANSEL(6) | ADC_SEQ_CTRL_CHANSEL(7) | ADC_SEQ_CTRL_CHANSEL(8) | ADC_SEQ_CTRL_CHANSEL(9) | ADC_SEQ_CTRL_CHANSEL(10) | ADC_SEQ_CTRL_CHANSEL(11) | ADC_SEQ_CTRL_BURST));
  Chip_ADC_StartBurstSequencer(LPC_ADC, ADC_SEQA_IDX);
  // Clear all pending interrupts
  Chip_ADC_ClearFlags(LPC_ADC, Chip_ADC_GetFlags(LPC_ADC));
  // Enable sequencer
  Chip_ADC_EnableSequencer(LPC_ADC, ADC_SEQA_IDX);
  Chip_ADC_StartSequencer(LPC_ADC, ADC_SEQA_IDX);
}

/*======================================================================*/
/* USART                                                                */
/*======================================================================*/
//-----------------------------
// タスクスイッチ励起
//-----------------------------
void us_csw(void) {
  taskYIELD();
}

//-----------------------------
// 初期化

//-----------------------------
static uint32_t us_init(uint32_t baud) {
//  usart0_dma_csw = us_csw;
  return usart0_dma_init(baud, txd, sizeof(txd), rxd, sizeof(rxd));
}

//-----------------------------
// 終了
//-----------------------------
static void us_deinit(void) { }

/*======================================================================*/
/* DX Communication task                                                */
/*======================================================================*/
//-----------------------------
// アイテムのレンジチェック(個別)
//-----------------------------
bool ValidPrevMapRange(const void *pmem, uint8_t u_st, uint8_t u_len) {
  bool result = true;
  const TE097CtrlTable *pNewControlTable = pmem;
  int i;

  for (i = 0; i < 12; i++) {
    if(IS_SAFE_VAR(pNewControlTable, pNewControlTable->PinCfg[i], u_st, u_len)) {
      if(pNewControlTable->PinCfg[i] > 20) result = false;
    }
  }
  if(IS_SAFE_VAR(pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if(pNewControlTable->OUT > 0xfff) result = false;
  }
  return result;
}

//-----------------------------
// アイテムのレンジチェック(一括)
//-----------------------------
bool ValidNextMapRange(const void *pmem) {
  const TE097CtrlTable *pNewControlTable = pmem;

  return(
    (pNewControlTable->PinCfg[0] > 20)
    || (pNewControlTable->PinCfg[1] > 20)
    || (pNewControlTable->PinCfg[2] > 20)
    || (pNewControlTable->PinCfg[3] > 20)
    || (pNewControlTable->PinCfg[4] > 20)
    || (pNewControlTable->PinCfg[5] > 20)
    || (pNewControlTable->PinCfg[6] > 20)
    || (pNewControlTable->PinCfg[7] > 20)
    || (pNewControlTable->PinCfg[8] > 20)
    || (pNewControlTable->PinCfg[9] > 20)
    || (pNewControlTable->PinCfg[10] > 20)
    || (pNewControlTable->PinCfg[11] > 20)
  );
}

//-----------------------------
// NVM初期化時
//-----------------------------
void InitNVMValue(void *pCtrl, const void *pNVM) {
  TE097CtrlTable *pControlTable = pCtrl;
  const TE097CtrlTable *pNVMdata = pNVM;
  int i;

  for(i = 0; i < 12; i++) {
    pControlTable->PinCfg[i] = pNVMdata->PinCfg[i];
    pin_config(i, pControlTable->PinCfg[i]);
  }
}

//-----------------------------
// 読込パケット時
//-----------------------------
void ReadAction(void *pCtrl) {
  TE097CtrlTable *pControlTable = pCtrl;
  int i;
  uint16_t w = 0;

  for(i = 0; i < 12; i++) {
    pControlTable->ADV[i] = ADC_DR_RESULT(Chip_ADC_GetDataReg(LPC_ADC, i));
    w |= (Chip_GPIO_GetPinState(LPC_GPIO_PORT, 0, PinInfo[i].No) ? (1 << i) : 0);
  }
  pControlTable->IN = w;
}

//-----------------------------
// 書込パケット時
//-----------------------------
void WriteAction(const void *pNewCtrl, void *pCtrl, bool force) {
  const TE097CtrlTable *pNewControlTable = pNewCtrl;
  TE097CtrlTable *pControlTable = pCtrl;
  int i;

  // LED On/Off
  if(pNewControlTable->LED != pControlTable->LED || force) {
    SetLED(pNewControlTable->LED);
  }
  // Pin Config
  for(i = 0; i < 12; i++) {
    if(pNewControlTable->PinCfg[i] != pControlTable->PinCfg[i] || force) {
      pin_config(i, pNewControlTable->PinCfg[i]);
    }
  }
  // OUT
  if(pNewControlTable->OUT != pControlTable->OUT || force) {
    pControlTable->OUT &= 0xfff;
    for(i = 0; i < 12; i++) Chip_GPIO_SetPinState(LPC_GPIO_PORT, 0, PinInfo[i].No, pNewControlTable->OUT & (1 << i) ? true : false);
  }
}

/*======================================================================*/
/* Stack over flow hook                                                 */
/*======================================================================*/
void vApplicationStackOverflowHook(TaskHandle_t pxTask, char *pcTaskName) {
  int i = 0;
  taskDISABLE_INTERRUPTS();
  while(1) {
    for(i = 0; i < 1000; i++) {
      SetLED(true);
      Delay_us (1000 - i);
      SetLED(false);
      Delay_us (i);
    ToggleLED();
    }
    Delay_ms(50);
    Chip_WWDT_Feed(LPC_WWDT);
  }
}

/*======================================================================*/
/* TASK                                                                 */
/*======================================================================*/
#ifdef _USE_MYTASK
void MyTask(void *pvParameters) {
  portTickType t;
  static uint16_t o=0;

  t = xTaskGetTickCount ();
  while(1) {
    if (++o>0xfff) o=0;
    DX_WriteMem(_ADDR_OUT, (uint8_t *)&o, 2);
    ToggleLED();
    vTaskDelayUntil (&t, 1);
  }
}
#endif

void TaskDXComm(void *pvParameters) {
  DX_ClientCommTask();
}

/*======================================================================*/
/* main                                                                 */
/*======================================================================*/
void wdt_feed (void) { Chip_WWDT_Feed(LPC_WWDT); }
void sys_reset (void) { NVIC_SystemReset(); }

TDXClient_ConfParam DX_ConfParam = {
  {(void *)&DEFAULT_DATA, sizeof(TE097CtrlTable), wdt_feed, sys_reset, 0x7F00, 256},
  { us_init, us_deinit, usart0_dma_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { ValidPrevMapRange, ValidNextMapRange, InitNVMValue, ReadAction, WriteAction },
};

void main(void) {
  Chip_SYSCTL_SetBODLevels(SYSCTL_BODRSTLVL_1, SYSCTL_BODINTVAL_LVL1);
  Chip_SYSCTL_EnableBODReset();

  PIO_Configure (pins, PIO_LISTSIZE (pins));
  init_wdt();
  init_adc();

  // DXクライアント機能設定
  if (DX_Config(&DX_ConfParam)) {
    xTaskCreate(TaskDXComm, NULL, 200, NULL, 1, NULL);
#ifdef _USE_MYTASK
    xTaskCreate(MyTask, NULL, 200, NULL, 1, NULL);
#endif
    vTaskStartScheduler();
  }
  while(1) {
    Chip_WWDT_Feed(LPC_WWDT);
    ToggleLED();
    Delay_ms(200);
  }
}
